@@ -1,4 +1,4 @@
-/* Copyright (c) 2021-2023 Advanced Micro Devices, Inc.
+/* Copyright (c) 2021-2024 Advanced Micro Devices, Inc.
 
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
@@ -47,6 +47,9 @@ enum elf_amdgpu_machine_t : uint32_t
   EF_AMDGPU_MACH_AMDGCN_GFX906 = 0x02f,
   EF_AMDGPU_MACH_AMDGCN_GFX908 = 0x030,
   EF_AMDGPU_MACH_AMDGCN_GFX90A = 0x03f,
+  EF_AMDGPU_MACH_AMDGCN_GFX940 = 0x040,
+  EF_AMDGPU_MACH_AMDGCN_GFX941 = 0x04b,
+  EF_AMDGPU_MACH_AMDGCN_GFX942 = 0x04c,
   EF_AMDGPU_MACH_AMDGCN_GFX1010 = 0x033,
   EF_AMDGPU_MACH_AMDGCN_GFX1011 = 0x034,
   EF_AMDGPU_MACH_AMDGCN_GFX1012 = 0x035,
@@ -56,6 +59,8 @@ enum elf_amdgpu_machine_t : uint32_t
   EF_AMDGPU_MACH_AMDGCN_GFX1100 = 0x041,
   EF_AMDGPU_MACH_AMDGCN_GFX1101 = 0x046,
   EF_AMDGPU_MACH_AMDGCN_GFX1102 = 0x047,
+  EF_AMDGPU_MACH_AMDGCN_GFX1200 = 0x048,
+  EF_AMDGPU_MACH_AMDGCN_GFX1201 = 0x04e,
 };
 
 using os_agent_id_t = uint32_t;
@@ -78,6 +83,8 @@ struct os_agent_info_t
   size_t max_waves_per_simd{ 0 };
   /* Number of shader engines.  */
   size_t shader_engine_count{ 0 };
+  /* Number of Compute Accelerator Chips.  */
+  uint32_t xcc_count{ 0 };
   /* PCI vendor id.  */
   uint32_t vendor_id{ 0 };
   /* PCI device id.  */
@@ -110,6 +117,8 @@ struct os_agent_info_t
   bool watchpoint_exclusive{ false };
   /* indicates if precise memory operations reporting is supported.  */
   bool precise_memory_supported{ false };
+  /* indicates if precise ALU operations reporting is supported.  */
+  bool precise_alu_exceptions_supported{ false };
   /* indicates that the command process firmware is supported.  */
   bool firmware_supported{ false };
   /* indicates that the trap temporaries are always set up.  */
@@ -145,7 +154,9 @@ enum class os_exception_code_t : uint32_t
   queue_wave_math_error = EC_QUEUE_WAVE_MATH_ERROR,
   queue_wave_illegal_instruction = EC_QUEUE_WAVE_ILLEGAL_INSTRUCTION,
   queue_wave_memory_violation = EC_QUEUE_WAVE_MEMORY_VIOLATION,
-  queue_wave_aperture_violation = EC_QUEUE_WAVE_APERTURE_VIOLATION,
+  queue_wave_address_error = /* FIXME: a future change to the ioctl will rename
+                                APERTURE_VIOLATION -> ADDRESS_ERROR.  */
+  EC_QUEUE_WAVE_APERTURE_VIOLATION,
   queue_packet_dispatch_dim_invalid = EC_QUEUE_PACKET_DISPATCH_DIM_INVALID,
   queue_packet_dispatch_group_segment_size_invalid
   = EC_QUEUE_PACKET_DISPATCH_GROUP_SEGMENT_SIZE_INVALID,
@@ -218,8 +229,10 @@ enum class os_exception_mask_t : uint64_t
   queue_wave_illegal_instruction
   = KFD_EC_MASK (EC_QUEUE_WAVE_ILLEGAL_INSTRUCTION),
   queue_wave_memory_violation = KFD_EC_MASK (EC_QUEUE_WAVE_MEMORY_VIOLATION),
-  queue_wave_aperture_violation
-  = KFD_EC_MASK (EC_QUEUE_WAVE_APERTURE_VIOLATION),
+  queue_wave_address_error
+  = KFD_EC_MASK (/* FIXME: a future change to the ioctl will rename
+       APERTURE_VIOLATION -> ADDRESS_ERROR.  */
+                 EC_QUEUE_WAVE_APERTURE_VIOLATION),
   queue_packet_dispatch_dim_invalid
   = KFD_EC_MASK (EC_QUEUE_PACKET_DISPATCH_DIM_INVALID),
   queue_packet_dispatch_group_segment_size_invalid
@@ -299,6 +312,8 @@ enum class os_wave_launch_trap_mask_t : uint32_t
   fp_inexact = KFD_DBG_TRAP_MASK_FP_INEXACT,
   int_divide_by_zero = KFD_DBG_TRAP_MASK_INT_DIVIDE_BY_ZERO,
   address_watch = KFD_DBG_TRAP_MASK_DBG_ADDRESS_WATCH,
+  wave_start = KFD_DBG_TRAP_MASK_TRAP_ON_WAVE_START,
+  wave_end = static_cast<uint32_t> (KFD_DBG_TRAP_MASK_TRAP_ON_WAVE_END)
 };
 template <> struct is_flag<os_wave_launch_trap_mask_t> : std::true_type
 {
@@ -350,6 +365,15 @@ enum class os_wave_launch_mode_t : uint32_t
   disable = 4,     /* Disable launching any new waves.  */
 };
 
+enum class os_process_flags_t : uint32_t
+{
+  precise_memory = 1 << 0,
+  precise_alu_exceptions = 1 << 1
+};
+template <> struct is_flag<os_process_flags_t> : std::true_type
+{
+};
+
 class os_driver_t
 {
 protected:
@@ -370,10 +394,17 @@ public:
   static std::unique_ptr<os_driver_t>
   create_driver (std::optional<amd_dbgapi_os_process_id_t> os_pid);
 
+  static std::unique_ptr<os_driver_t>
+  create_driver (const amd_dbgapi_core_state_data_t &core_state);
+
   virtual bool is_valid () const = 0;
 
   virtual amd_dbgapi_status_t check_version () const = 0;
 
+  virtual amd_dbgapi_status_t
+  create_core_state_note (const os_runtime_info_t &runtime_info,
+                          amd_dbgapi_core_state_data_t *data) const = 0;
+
   virtual amd_dbgapi_status_t
   agent_snapshot (os_agent_info_t *snapshots, size_t snapshot_count,
                   size_t *agent_count,
@@ -437,7 +468,8 @@ public:
     os_wave_launch_trap_mask_t *previous_value = nullptr,
     os_wave_launch_trap_mask_t *supported_mask = nullptr) const = 0;
 
-  virtual amd_dbgapi_status_t set_precise_memory (bool enabled) const = 0;
+  virtual amd_dbgapi_status_t
+  set_process_flags (os_process_flags_t flags) const = 0;
 
   virtual amd_dbgapi_status_t
   xfer_global_memory_partial (amd_dbgapi_global_address_t address, void *read,
@@ -445,8 +477,10 @@ public:
 };
 
 template <> std::string to_string (os_agent_info_t os_agent_info);
+template <> std::string to_string (kfd_dbg_device_info_entry);
 template <> std::string to_string (os_exception_code_t exception_code);
 template <> std::string to_string (os_exception_mask_t exception_mask);
+template <> std::string to_string (os_process_flags_t flags);
 template <> std::string to_string (os_queue_snapshot_entry_t snapshot);
 template <> std::string to_string (os_runtime_info_t runtime_info);
 template <> std::string to_string (os_runtime_state_t runtime_state);
